      subroutine compchol(task,a,b,dim,ifdrv,dbda,db,da,status)
 
c     Subroutine to
c     Task = 1: compute the Cholesky decomposition of a symmetric
c               positive definite matrix A.  The algorithm is based
c               on subroutine dpoda as written on August 1, 1982 by
c               C.B. Moler as part of the linpack routines.
c     Task = 2: unravel derivatives wrt B into wrt A.
c
c     A     [in] symmetric positive definite matrix A(dim,dim).
c     B    [out] upper triangle of B(dim,dim) contains the Cholesky
c                decomposition such that B'B = A.  The lower triangle
c                contains zeroes.  Note: A may B may share the same
c                memory space.  A and B are not used if task=2.
c     ifdrv [in] logical variable for whether derivatives dAdB are
c                desired.  Note: if later on dA needs to be found
c                from dB, dBdA is required and ifdrv should therefore
c                be on.
c     dBdA [out/in] Derivatives of B wrt A: dBdA(i,j,k,l) contains the
c                derivative of B(i,j) wrt A(k,l).  Many elements are
c                zero, mainly if i>j or k>l or k>i or l>j.
c                NOTE: A is always symmetric and the derivatives of B
c                wrt a(k,l) are therefore equal to wrt a(l,k).  dBdA
c                stores these derivatives collapsed (accumulated) onto
c                the upper diagonal.  In other words, dBdA(.,.,k,l)
c                is not symmetric but upper triangular.
c                dBdA is output for task=1 and ifdrv=.true. and input
c                for task=2.
c     dB    [in] Derivatives wrt B.  (not used if task=1)
c     dA   [out] Derivatives wrt A.  (not used if task=1)
c                NOTE CAREFULLY: the entire matrix dA is formed, not
C                just the upper triangle.  The derivative wrt covariance
c                element (k,l), for instance, is therefore the sum of
c                dA(k,l) and dA(l,k).
c
c     Error return is used if task=1 and A is not positive definite.
c     There is no check that A is symmetric; only its upper triangle
c     is used.
c
c     NOTE: The formulae for Cholesky decomposition are as follows:
c     b11 = sqrt(a11)
c     b1. = a1./b11
c     b22 = sqrt(a22-b12*b12)
c     b2. = (a2.-b12*b1.)/b22
c     b33 = sqrt(a33-b13*b13-b23*b23)
c     b3. = (a3.-b13*b1.-b23*b2.)/b33
c     b44 = sqrt(a44-b14*b14-b24*b24-b34*b34)
c     b4. = (a4.-b14*b1.-b24*b2.-b34*b3.)/b44
c     et cetera; see E. Bodewig, "Matrix Calculus", 1959, page 127.
c
c     Constantijn Panis,  10 September 1991
 
      implicit none
      integer task,dim,dim2,status,i,j,k,l
      double precision a(dim,dim),b(dim,dim),dbda(dim,dim,dim,dim),
     *   db(dim,dim),da(dim,dim),t,s,dada,innerp
      logical ifdrv
 
c     Compute Cholesky decomposition or unravel derivatives?
      if(task.eq.2)goto 200
 
c     Compute Cholesky decomposition:
  100 status=0
      if(ifdrv)call zerodp(dim**4,dbda)
      do 110 j=1,dim
         s = a(j,j)
         do 120 i=1,j-1
            t = (a(i,j)-innerp(b(1,i),b(1,j),i-1))/b(i,i)
            b(i,j) = t
            b(j,i) = 0.d0
            s = s - t*t
            if(.not.ifdrv)goto 120
c           Derivatives of b(i,j) wrt a(k,l):
            do 130 l=1,j
               do 135 k=1,min(i,l)
                  if(i.eq.k.and.j.eq.l)then
                     dada = 1.d0
                  else
                     dada = 0.d0
                  endif
                  dbda(i,j,k,l) = (dada -
     *               innerp(dbda(1,i,k,l),b(1,j),i-1) -
     *               innerp(dbda(1,j,k,l),b(1,i),i-1) -
     *               b(i,j)*dbda(i,i,k,l)) / b(i,i)
  135          continue
  130       continue
  120    continue
         if(s.le.0.d0)goto 800
         b(j,j) = dsqrt(s)
         if(.not.ifdrv)goto 110
c        Derivatives of b(j,j) wrt a(k,l):
         do 140 l=1,j
            do 150 k=1,min(i,l)
               if(j.eq.k.and.j.eq.l)then
                  dada = .5d0
               else
                  dada = 0.d0
               endif
               dbda(j,j,k,l) = (dada -
     *            innerp(b(1,j),dbda(1,j,k,l),j-1)) / b(j,j)
  150       continue
  140    continue
  110 continue
      goto 999
 
c     Unravel derivatives:
  200 dim2=dim*dim
      do 210 l=1,dim
         da(l,l) = innerp(db,dbda(1,1,l,l),dim2)
         do 220 k=1,l-1
            da(k,l) = .5d0*innerp(db,dbda(1,1,k,l),dim2)
            da(l,k) = da(k,l)
  220    continue
  210 continue
      goto 999
 
  800 status=1
  999 return
      end
