      subroutine cov2corr(task,k,covar,corr,dcorr,dcovar,status)
 
c     Subroutine to convert covariance matrix into correlation matrix
c     with standard deviations on the diagonal.  If the covariance
c     matrix is negative definite (and thus illegal), the alternate
c     return is used.
c     There is no check that the input matrix is symmetric.
c     Arguments:
c     task   [in]  = 1: compute correlation matrix
c                  = 2: unravel derivatives
c     k      [in]  dimension of sigma
c     covar  [in]  covariance matrix
c     corr   [out] correlation matrix
c     dcorr  [in]  derivatives wrt correlation matrix
c     dcovar [out] derivatives wrt covariance matrix.
c     status [out] 0 = all OK
c                 -1 = matrix singular
c                  1 = invalid covariance matrix
c
c     Detection of singular/invalid covariance matrices is based solely on
c     simple correlations being inside [-1,1].  Multivariate covariance
c     matrices may be invalid and this routine would still return status=0.
c     Also see subroutine ensposdef.
c
c     Note: covar and corr may share the same memory space, as may
c        dcovar and dcorr.  If task=1, dcovar and dcorr are not used;
c        if task=2, covar is not used, but corr is.
c
c     BUGS: There is a bug in the unraveling part of this routine.
c           If a standard deviation is zero, the derivative wrt the
c           variance is returned as zero, rather than some limit case.
c
c     Constantijn Panis, 11 April 2003
c     Last change: Determine whether correlation is +/-1 in way that is more
c     robust to floating point inaccuracies.  Also introduce status=-1.
 
      implicit none
      integer task,k,status,i,j,l
      double precision covar(k,k),corr(k,k),dcorr(k,k),dcovar(k,k),
     *   deriv,eps,temp
c     Numerical inaccuracy tolerance for correlations:
      parameter(eps=1.d-13)
 
      status=0
 
c     compute correlation matrix or unravel derivatives?
      if(task.eq.2)goto 200
 
c     Compute correlation matrix:
      do 110 i=1,k
         if(covar(i,i).le.0.d0)then
c           Variance is zero or negative
            if(covar(i,i).lt.-eps)goto 800
c           Variance is zero; singular matrix:
            status=-1
            corr(i,i)=0.d0
c           Check that associated covariances are zero:
            do 140 j=1,k
               if(abs(covar(i,j)).gt.eps)goto 800
               corr(i,j)=0.d0
               corr(j,i)=0.d0
  140       continue
         else
c           variance is positive (normal case):
            corr(i,i)=dsqrt(covar(i,i))
            do 150 j=1,i-1
c              if corr(j,j) is zero, corr(i,j) has already been set to
c              zero in loop 140.
               if(corr(j,j).eq.0.d0)goto 150
               corr(i,j) = covar(i,j)/(corr(i,i)*corr(j,j))
               if(abs(corr(i,j)).ge.1.d0)then
c                 corr=-1/1 indicates singularity.  Allow for numerical inaccura
c                 (E.g., Trouble\trouble13)
                  if(abs(corr(i,j)).eq.1.d0)then
                     status=-1
                  elseif(abs(corr(i,j)).lt.1.d0+eps)then
                     if(corr(i,j).lt.0.d0)corr(i,j)=-1.d0
                     if(corr(i,j).gt.0.d0)corr(i,j)=1.d0
                     status=-1
                  else
                     goto 800
                  endif
               endif
               corr(j,i)=corr(i,j)
  150       continue
         endif
  110 continue
      goto 999
 
c     Unravel derivatives.  Note that since covar and corr may
c     share the same space, they both may be the correlation matrix.
c     Account for the fact that dcorr may not be symmetric.
c     Note that the derivative wrt a variance involves not only the
c     derivative wrt a standard deviation, but also wrt all correlation
c     coefficients associated with that variance.  Note that the
c     derivative of rho12 wrt sig11 is -.5*rho12/sig11.
c     First unravel derivatives wrt standard errors into wrt variances;
c     then do wrt correlations into wrt covariances.  The split is needed
c     to be able to let dcovar and dcorr share memory space.
  200 do 210 i=1,k
         temp=corr(i,i)*corr(i,i)
c        This is a bug:  really should be some limit case if temp=0.
         if(temp.ne.0.d0)temp = 1.d0/temp
         deriv = -corr(i,i)*dcorr(i,i)
         do 220 l=1,k
            if(i.ne.l)deriv=deriv+corr(i,l)*(dcorr(i,l)+dcorr(l,i))
  220    continue
         dcovar(i,i) = -.5d0*temp*deriv
  210 continue
      do 230 i=1,k
         do 240 j=1,k
            if(j.eq.i)goto 240
            temp = corr(i,i)*corr(j,j)
c           This is a bug:  really should be some limit case if temp=0.
            if(temp.ne.0.d0)temp = 1.d0/temp
            dcovar(i,j) = temp*dcorr(i,j)
  240    continue
  230 continue
      goto 999
 
c     Correlation is outside [-1,1]
  800 status=1
  999 return
      end
