      subroutine fillarma(task,p,q,nt,phi,vareps,theta,covar,time,
     *   dcovar,dphi,dvareps,dtheta,ipt,jpt)
 
c     Subroutine to fill out the covariance matrix of an ARMA(p,q)
c     process:
c
c     y(t) = phi(1)*y(t-1) + ... + phi(p)*y(t-p) +
c            eps(t) + theta(1)*eps(t-1) + ... + theta(q)*eps(t-q)
c
c     Arguments:
c        task    [in]  1=fill out covariance matrix;   2=unravel derivatives
c        p,q     [in]  ARMA(p,q) order
c        nt      [in]  number of (integer-valued) time periods
c        phi     [in]  autoregressive parameters
c        vareps  [in]  variance of innovation term
c        theta   [in]  moving average parameters
c        covar   [out] covariance matrix (assigned, not accumulated!)
c        time    [in]  (integer-valued) time values (not necessarily contiguous)
c                      Note:  if time(i)=NaN, the i-th row and column of the
c                      covariance matrix will be zero.  ARMA residuals apply to
c                      either all or none continuous outcomes of the level
c                      within which there is autocorrelation.  The implication
c                      is that time will never be NaN in the current routine.
c        dcovar  [in]  derivatives wrt covariance matrix
c        dphi    [out] derivatives wrt autoregressive parameters
c        dvareps [out] derivatives wrt variance of innovation term
c        dtheta  [out] derivatives wrt moving average parameters
c
c     Note: dphi, dvareps, and dtheta are accummulated in this routine,
c     not assigned.
c
c     The error return is used in case the parameterization is such that
c     the process is nonstationary.  The error return is also used if
c     p or q exceed maxdim.
c
c     BUG:  not all nonstationarities may be caught.
c
c     Constantijn Panis, 26 August 1996
c     Last change:  check for time(.)=NaN removed.
 
      implicit none
      integer maxdim
      parameter(maxdim=9)
      integer task,p,q,nt,ipt,jpt,maxpq,maxpq1,dimg
      double precision phi(p),vareps,theta(q),covar(nt,nt),
     *   dcovar(nt,nt),dphi(p),dvareps,dtheta(2),eye(0:maxdim-1),
     *   a(maxdim,maxdim),b(maxdim+1),g(1-maxdim:maxdim),
     *   ai(maxdim,maxdim),da(maxdim,maxdim),deye(0:maxdim-1),
     *   dai(maxdim,maxdim),db(maxdim+1),dg(0:maxdim)
      real time(nt)
 
c     Extra step to get dimensions of a, etc, right:
      if(p.gt.maxdim.or.q.gt.maxdim)goto 800
      dimg=max(0,p-1)
      maxpq=max(p,q)
      maxpq1=max(p,q+1)
      call fillarma_(task,p,q,nt,maxpq,maxpq1,dimg,phi,vareps,
     *   theta,covar,time,dcovar,dphi,dvareps,dtheta,
     *   eye,a,b,g,ai,deye,da,dai,db,dg,ipt,jpt)
      goto 999
 
c     Diagnosis:  maxdim is exceeded
  800 if(ipt.gt.0)call writeaia(8,72,1,'ERROR:  the maximum '//
     *   'autoregressive or moving average order of an ARMA '//
     *   'process (',maxdim,') is exceeded.')
      if(jpt.gt.0)call writeaia(6,72,1,'ERROR:  the maximum '//
     *   'autoregressive or moving average order of an ARMA '//
     *   'process (',maxdim,') is exceeded.')
      stop
  999 return
      end
c==========================================================================
      subroutine fillarma_(task,p,q,nt,maxpq,maxpq1,dimg,phi,vareps,
     *   theta,covar,time,dcovar,dphi,dvareps,dtheta,
     *   eye,a,b,g,ai,deye,da,dai,db,dg,ipt,jpt)
      implicit none
      integer task,p,q,nt,ipt,jpt,maxpq,maxpq1,dimg,i,j,l,lagmax,
     *   status,first,num,lnz,lastnonzero
      double precision phi(p),vareps,theta(q),covar(nt,nt),
     *   dcovar(nt,nt),dphi(p),dvareps,dtheta(q),
     *   gam,dgam,dprowsum,innerp,innerpr,ginnerp
      double precision eye(0:maxpq-1),a(p,p),b(maxpq1),ai(p,p),
     *   deye(0:maxpq-1),da(p,p),dai(p,p),db(maxpq1),g(-dimg:maxpq1-1),
     *   dg(0:maxpq1-1)
      real time(nt),timemin,timemax,realvecmin,realvecmax
 
c     First compute anything that is needed for both computing the
c     covariance matrix and unraveling derivatives.  Begin by determining
c     the largest time lag, lagmax:
      timemin=realvecmin(time,nt)
      timemax=realvecmax(time,nt)
c     Time marks are guaranteed (by figure_res) to be integer-valued.
      lagmax=int(timemax-timemin)
 
      if(p.eq.0.and.q.eq.0)then
c        ARMA(0,0):  just an iid residual
         g(0)=vareps
      elseif(p.eq.0)then
c        MA(q):   y(t) = e(t) + SUM(theta(i)*e(t-i))
         g(0)=1+innerp(theta,theta,q)
         do 110 l=1,q
            g(l)=theta(l)
            do 115 i=1,q-l
               g(l)=g(l)+theta(i)*theta(i+l)
  115       continue
  110    continue
         call mult(g(0),vareps,g(0),q+1)
c        gamma(l)=0, l>q
      else
c        AR(p):      y(t) = SUM(phi(i)*y(t-i)) + e(t)
c        ARMA(p,q):  y(t) = SUM(phi(i)*y(t-i)) + e(t) + SUM(theta(i)*e(t-i))
c        Compute A from autocovariance 0 through p-1:
         call zerodp(p*p,a)
         a(1,1)=1-innerp(phi,phi,p)
         do 120 i=2,p
            a(1,i)=-2*innerp(phi,phi(i),p-i+1)
  120    continue
         do 130 l=1,p-1
            a(1+l,1+l)=1
            do 135 i=1,p
               a(1+l,1+abs(i-l))=a(1+l,1+abs(i-l))-phi(i)
  135       continue
  130    continue
c        Compute B from autocovariance 0 via p-1 all the way to maxpq1-1:
         call zerodp(maxpq1,b)
         if(q.eq.0)then
c           AR(p):  Compute B (only zero-th element is nonzero):
            b(1)=1.d0
         else
c           ARMA(p,q):  y(t) = SUM(phi(i)*y(t-i)) + e(t) + SUM(theta(i)*e(t-i))
c           First compute E[y(t)e(t-i)], i=0,maxpq-1  (omitting vareps):
            call zerodp(maxpq,eye)
            eye(0)=1
            do 140 i=1,maxpq-1
               if(i.le.p)eye(i)=eye(i)+phi(i)
               if(i.le.q)eye(i)=eye(i)+theta(i)
               do 145 j=1,min(p,i-1)
                  eye(i)=eye(i)+phi(j)*eye(i-j)
  145          continue
  140       continue
            b(1)=1+innerp(theta,theta,q)
            do 150 i=1,min(p,q)
               b(1)=b(1)+2*phi(i)*innerp(theta(i),eye,q-i+1)
  150       continue
            do 160 l=2,q+1
               b(l)=innerp(theta(l-1),eye,q-l+2)
  160       continue
         endif
c        Compute g(0) through g(p-1) as vareps*Ai*B:
         status=0
         call nrminv(1,a,ai,p,0.d0,0.d0,status)
         if(status.gt.0)goto 800
         call m2mult(000,p,p,1,ai,b,g(0),0)
         call mult(g(0),vareps,g(0),p)
c        Number of nontrivial autocovariances is maxpq1 = max(p,q+1) --
c        g(0) through g(maxpq1-1).
         do 170 i=p,maxpq1-1
            g(i) = innerpr(phi,g(i-p),p) + b(i+1)*vareps
  170    continue
c        For l>=maxpq1, gamma(l) = SUM(phi(i)*gamma(l-i))
      endif
c     Nonstationary processes result in error return
      if(g(0).le.0.d0)goto 810
 
c     Fill out covariance matrix or unravel derivatives?
      if(task.eq.2)goto 300
 
c     Fill out covariance matrix:
  200 call zerodp(nt*nt,covar)
c     Diagonal elements:
      if(p.eq.0.and.q.eq.0)then
c        ARMA(0,0):  just an iid residual
         call putrow(g(0),covar,nt,nt+1)
      elseif(p.eq.0)then
c        MA(q):   y(t) = e(t) + SUM(theta(i)*e(t-i))
c        Diagonal elements:
         do 220 l=0,min(lagmax,q)
            call armaonecov(1,l,nt,covar,g(l),dgam,dcovar,time)
  220    continue
      else
c        AR(p):      y(t) = SUM(phi(i)*y(t-i)) + e(t)
c        ARMA(p,q):  y(t) = SUM(phi(i)*y(t-i)) + e(t) + SUM(theta(i)*e(t-i))
         do 230 l=0,lagmax
            if(l.lt.maxpq1)then
               gam=g(l)
            else
c              gam(l) = phi(1)*gam(l-1) + phi(2)*gam(l-2) + ...
               gam=innerpr(phi,g(0),p)
            endif
c           Push gammas one level down, place gam(l) in last element
            if(l.ge.p)call pushdown8(p,g(0),gam)
            call armaonecov(1,l,nt,covar,gam,dgam,dcovar,time)
  230    continue
      endif
      call reflect(covar,nt)
      goto 999
 
c     Unravel derivatives:
  300 call zerodp(maxpq1,dg)
      if(p.eq.0.and.q.eq.0)then
c        ARMA(0,0):  just an iid residual
         dvareps=dvareps+dprowsum(dcovar,nt+1,nt)
      elseif(p.eq.0)then
c        MA(q):   y(t) = e(t) + SUM(theta(i)*e(t-i))
         do 310 l=min(lagmax,q),0,-1
            call armaonecov(2,l,nt,covar,g(l),dg(l),dcovar,time)
  310    continue
c        Unravel wrt gammas into wrt vareps:
         dvareps=dvareps+innerp(dg,g(0),q+1)/vareps
c        Unravel wrt gammas into wrt theta:
         call mult(dg,vareps,dg,q+1)
         call add(theta,dtheta,2*dg(0),q)
         do 320 l=1,q
            dtheta(l)=dtheta(l)+dg(l)
            do 325 i=1,q-l
               dtheta(i)=dtheta(i)+theta(i+l)*dg(l)
               dtheta(i+l)=dtheta(i+l)+theta(i)*dg(l)
  325       continue
  320    continue
      else
c        AR(p):      y(t) = SUM(phi(i)*y(t-i)) + e(t)
c        ARMA(p,q):  y(t) = SUM(phi(i)*y(t-i)) + e(t) + SUM(theta(i)*e(t-i))
c        There's no need to store all gammas; just store last p and
c        compute others on the fly, as needed.
         if(p.eq.1.and.q.eq.0.and.phi(1).ne.0.d0)then
c           AR(1) -- treat separately to prevent overhead
            g(0)=g(0)*phi(1)**lagmax
         else
c           Compute gammas and store gam(lagmax) in g(0)
            if(lagmax.lt.p)then
               call copy8(g(0),g(-lagmax),p)
            else
               call copy8(g(0),g(1-p),p)
               do 330 l=p,lagmax
                  if(l.lt.maxpq1)then
                     gam=g(l)
                  else
c                    gam(l) = SUM(phi(i)*gam(l-i))
                     gam=innerpr(phi,g(1-p),p)
                  endif
c                 Push gammas one position down, place gam(l) in g(0)
                  call pushdown8(p,g(1-p),gam)
  330          continue
            endif
         endif
c        Autocovariances will be recomputed below.  In principle,
c        autocovariance l is computed off the formula for l+p, except
c        where phi(p)=0, in which case the formula for l+p-1 applies,
c        et cetera.  Denote lnz as "last non-zero" phi:
         lnz=lastnonzero(p,phi)
c        g now contains                              and will below contain:
c              g(1-p) = gam(lagmax+1-p)                gam(lagmax+1-p)
c                     :                                   :
c              g(-1)  = gam(lagmax-1)                  gam(lagmax-1)
c              g(0)   = gam(lagmax)                    gam(l)
c              g(1)   = N/A                            gam(l+1)
c                     :                                   :
c              g(p-1) = N/A                            gam(l+p-1)
c              g(p)   = gam(p)                         gam(p)
c              g(p+1) = gam(p+1)                       gam(p+1)
c                     :                                   :
c              g(maxpq1-1) = gam(maxpq-1)              gam(maxpq-1)
         call zerodp(maxpq1,db)
         do 340 l=lagmax,0,-1
            call armaonecov(2,l,nt,covar,gam,dgam,dcovar,time)
c           Higher autocovariances contribute to this derivative, but
c           only autocovariances p through lagmax:
            first=max(1,p-l)
            num=min(p,lagmax-l)-first+1
            dgam=dgam+innerp(phi(first),dg(first-1),num)
c           Determine gam, push g up, replace g(0) by gam:
            if(l.ge.lagmax+1-p)then
               gam=g(l-lagmax)
            elseif(l.ge.maxpq1)then
c              gam(l) may be determined from the equation for gam(l+p), or
c              more generally, gam(l+lnz).  If all phi=0, gam does not change.
               if(lnz.gt.0)gam=
     *            (g(lnz-1)-innerpr(phi,g(0),lnz-1))/phi(lnz)
            elseif(l.lt.p)then
c              gam(l+p) is nontrivial, as is gam(l).  Note that gam(p)
c              through gam(maxpq1-1) are readily available.  By contrast,
c              gam(0) through gam(p-1) may have to be recomputed.
               gam=vareps*ginnerp(p,p,1,ai(l+1,1),b)
            else
c              g(p) through g(maxpq1-1) remain available:
               gam=g(l)
            endif
            call pushup8(p,g(0),gam)
c           dphi(i) =+ gam(l)*dgam(l+i)
            call add(dg(first-1),dphi(first),gam,num)
c           Push dg one position up, replacing element 0 by dg(l)
            call pushup8(p,dg,dgam)
            if(l.ge.p.and.l.lt.maxpq1)then
c              Take account of nontrivialities of g(p) through g(maxpq1-1):
               db(l+1)=dgam*vareps
               dvareps=dvareps+dgam*b(l+1)
            endif
  340    continue
 
c        Unravel wrt gammas into wrt vareps:
         dvareps=dvareps+innerp(dg,g(0),p)/vareps
 
c        Unravel wrt g(0) through g(p-1) into wrt A
         call mult(dg,vareps,dg,p)
         call m2mult(010,p,1,p,dg,b,dai,0)
         call zerodp(p*p,da)
         call m3mult(1010,p,p,p,p,ai,dai,ai,da,-1)
c        Unravel wrt A into wrt phi:
         call add(phi,dphi,-2*da(1,1),p)
         do 350 i=2,p
            call add(phi(i),dphi,-2*da(1,i),p-i+1)
            call add(phi,dphi(i),-2*da(1,i),p-i+1)
  350    continue
         do 360 l=1,p-1
            do 365 i=1,p
               dphi(i)=dphi(i)-da(1+l,1+abs(i-l))
  365       continue
  360    continue
 
         if(q.gt.0)then
c           Unravel wrt g(0) through g(p-1) into wrt B; note that wrt g(p)
c           through g(maxpq-1) have already been unraveled into wrt B.
            call m2mult(100,p,p,1,ai,dg,db,0)
c           Unravel wrt B into wrt eye, phi, and theta
            call add(theta,dtheta,2*db(1),q)
            call zerodp(maxpq,deye)
            do 370 i=1,min(p,q)
               dphi(i)=dphi(i)+2*db(1)*innerp(theta(i),eye,q-i+1)
               call add(eye,dtheta(i),2*phi(i)*db(1),q-i+1)
               call add(theta(i),deye,2*phi(i)*db(1),q-i+1)
  370       continue
            do 380 l=2,q+1
               call add(eye,dtheta(l-1),db(l),q-l+2)
               call add(theta(l-1),deye,db(l),q-l+2)
  380       continue
c           Unravel wrt eye into wrt phi and theta:
            do 390 i=maxpq-1,1,-1
               if(i.lt.maxpq-1)deye(i)=deye(i)+
     *            innerp(phi,deye(i+1),min(p,maxpq-1-i))
               if(i.le.p)dphi(i)=dphi(i)+deye(i)
               if(i.le.q)dtheta(i)=dtheta(i)+deye(i)
               do 395 j=1,min(p,i-1)
                  dphi(j)=dphi(j)+deye(i)*eye(i-j)
  395          continue
  390       continue
         endif
      endif
      goto 999
 
c     Matrix A is singular
  800 if(ipt.gt.0)call writea2ia(8,72,1,'ERROR:  invalid AR or ARMA '//
     *   'process.  The system of ',p,' equations to solve for '//
     *   'autocovariances 0 through ',p-1,' has multiple solutions.')
      if(jpt.gt.0)call writea2ia(6,72,1,'ERROR:  invalid AR or ARMA '//
     *   'process.  The system of ',p,' equations to solve for '//
     *   'autocovariances 0 through ',p-1,' has multiple solutions.')
      stop
c     ARMA process is nonstationary
  810 if(ipt.gt.0)write(8,811)
      if(jpt.gt.0)write(6,811)
  811 format(/,'ERROR:  nonstationary AR or ARMA process.')
      stop
  999 return
      end
c==========================================================================
      subroutine armaonecov(task,l,nt,covar,gam,dgam,dcovar,time)
 
c     Subroutine to fill in the covariance matrix of an ARMA process
c     with covariances with lag l, where l>=0.  Only the lower triangle
c     is filled.  If task=2, the derivatives are accumulated.
c
c     This routine is only called by fillarma.
c
c     Constantijn Panis, 26 August 1996
c     Last change: check for time()=NaN removed.
 
      implicit none
      integer task,l,nt,i,j
      double precision covar(nt,nt),gam,dcovar(nt,nt),dgam,dprowsum
      real time(nt)
 
      if(task.eq.1)then
         if(l.eq.0)then
            call putrow(gam,covar,nt,nt+1)
         else
            do 110 i=2,nt
               do 120 j=max(1,i-l),i-1
                  if(time(i)-time(j).eq.real(l))covar(i,j)=gam
  120          continue
  110       continue
         endif
      else
         if(l.eq.0)then
            dgam=dprowsum(dcovar,nt+1,nt)
         else
            dgam=0.d0
            do 210 i=2,nt
               do 220 j=max(1,i-l),i-1
                  if(time(i)-time(j).eq.real(l))dgam=dgam+
     *               dcovar(i,j)+dcovar(j,i)
  220          continue
  210       continue
         endif
      endif
 
      return
      end
 
