      subroutine m3mult0030(task,nrowa,nrowb,numblockc,dimblockc,
     *   a,b,cblock,d,factor)
 
c     Subroutine to compute A*B*C = D:
c
c            A(nrowa,nrowb)   B(nrowb,dimc)   C(dimc,dimc)   D(nrowa,dimc)
c
c     where dimc = numblockc*dimblockc
c
c     Matrix types can be:
c        0 = generic
c        1 = identity (Only its dimension is stored.)
c        2 = symmetric
c        3 = symmetric and block-diagonal (Only its first block, its
c            dimension, and number of blocks are stored.)
c        4 = sparse in the sense that there is at most one nonzero
c            element in any one row.  (Only its nonzero elements
c            and their column position are stored.)
c
c     This routine assumes that C is symmetric and block-diagonal.
c
c     The four matrices may be in transposed format.  Integer task
c     may be 1000,0100,0010, et cetera, depending on which matrices
c     are transposed.  The first digit refers to A; the second to B;
c     the third to C; the fourth to D.  For example,
c        task = 1100 means that A and B are transposed: A'*B'*C = D
c        task =   10 means that C is transposed: A*B*C' = D
c
c     Dimensions as given above assume no transpositions; if a matrix
c     is transposed, its dimensions really are reversed.
c
c     Integer factor determines:
c        factor = 0     D = A*B*C
c              else     D = D + factor*A*B*C
c
c     Constantijn Panis, 9 August 2002
c     New in version 3.00: c is symmetric and not expanded.  Conceptually,
c     it is block-diagonal with numblockc blocks of size (dimblockc,dimblockc).
c     Technically, only the first (dimblockc,dimblockc) block is passed in.
 
      implicit none
      integer task,nrowa,nrowb,numblockc,dimblockc,factor,i,j,k,l,
     *   ia,ib,id,ja,jb,jd,sa,sb,sd,jsa,jsb,jsd,cic,dimc
      double precision a(nrowa*nrowb),b(nrowb*numblockc*dimblockc),
     *   cblock(dimblockc*dimblockc),d(nrowa*numblockc*dimblockc),abij
 
      dimc=numblockc*dimblockc
 
c     In its most basic form, D = D + A*B*C, the routine is very compact:
c
c     do 110 i=1,nrowa
c        do 120 j=1,dimc
c           abij = 0.d0
c           do 130 k=1,nrowb
c 130          abij = abij + a(i,k)*b(k,j)
c           do 140 l=1,dimc
c 140          d(i,l) = d(i,l) + abij*c(j,l)
c 120    continue
c 110 continue
c
c     The 130 loop computes element (i,j) of A*B, hence the name abij.
c     This element is computed with every element of row j of C and
c     added to D(i,.).
c
c     Complications arise because this routine allows for A, B, C and D to
c     be stored in transposed form.  Note that A and B are only used in the
c     130 loop, and C and D are only used in the 140 loop.  The positions
c     that are used are ia, ib, ic and id.  In the 130 and 140 loops, they
c     increase by ja, jb, jc and jd, where j stands for "jump".  If A, B,
c     C and D are not-transposed, these jumps are nrowa, 1, nrowc and nrowd,
c     respectively, where nrow stands for the number of rows.  If A is
c     transposed, this changes to 1; if B is transposed, it becomes nrowb,
c     et cetera.  The first time an element is accessed in the 130 and 140
c     loops, we need A(i,1), B(1,j), C(j,1) and D(i,1); these are coded
c     as sa, sb, sc and sd, where s stands for "start".  These start off at
c     1 in all four cases, the first time the 130/140 loops are used.
c     The next time they increase by 1, nrowb, 1 and 1; these are coded
c     as jsa, jsb, jsc and jsd, where js stands for "jump in start".
c     If a matrix is transposed, the jumps are different.
 
      if(mod(task/1000,10).eq.0)then
         ja=nrowa
         jsa=1
      else
         ja=1
         jsa=nrowb
      endif
      if(mod(task/100,10).eq.0)then
         jb=1
         jsb=nrowb
      else
         jb=dimc
         jsb=1
      endif
c     C is symmetric
      if(mod(task,10).eq.0)then
         jd=nrowa
         jsd=1
      else
         jd=1
         jsd=dimc
      endif
 
      if(factor.eq.0)call zerodp(nrowa*dimc,d)
      sa=1
      sd=1
      do 110 i=1,nrowa
         sb=1
         do 120 j=1,dimc
            abij = 0.d0
            ia=sa
            ib=sb
            do 130 k=1,nrowb
               abij = abij + a(ia)*b(ib)
               ia=ia+ja
               ib=ib+jb
  130       continue
c           Loop over l=1,dimc and multiply AB(i,j) with C(j,.)
c           to accumulate D(i,.).  No need to cover the entire
c           C(j,.) because C is block-diagonal.
            cic=mod(j-1,dimblockc)+1
            id=sd+((j-1)/dimblockc)*dimblockc*jd
            if(factor.ne.0.and.factor.ne.1)abij=dble(factor)*abij
            do 140 l=1,dimblockc
               d(id) = d(id) + abij*cblock(cic)
               id=id+jd
               cic=cic+dimblockc
  140       continue
            sb=sb+jsb
  120    continue
         sa=sa+jsa
         sd=sd+jsd
  110 continue
 
      return
      end
 
