      subroutine metaparse(infile,macrofile,outunit)
 
c     Subroutine to strip comments out of a file and to expand macros.
c     Comments are characters delimited by /* and */.  Nested comments
c     are handled properly.
c     Arguments:
c        infile [in] character string with name of input file.  The
c                    file is closed upon return.
c        macrofile [in] character string to which expanded macro file
c                       needs to be written.  If macrofile=' ', the
c                       macro file is conneced to a scratch file.
c        outunit [in] integer unit to which macro file must be connected.
c                       If this unit is already open, an error occurs.
c
c     Embedded macros are resolved upon their invocation.  E.g.,
c        %MACRO race;  race=3  %MEND;
c        %MACRO covariates;
c           thisvar=1
c           thatvar=2
c           %race
c        %MEND;
c        %MACRO race;  race=1  %MEND;
c     %covariates now expands to
c           thisvar=1
c           thatvar=2
c           race=1
c
c     The output file will not have lines over 80 characters.
c     Spaces are replaced by line feeds if wrapping is needed.
c     The output file will also not contain blank lines.
c     Horizontal tabs are replaced by a single space.
c     Multiple spaces are reduced to just one space, except in strings
c        that are delimited by double quotes.
c     Leading and trailing spaces are omitted.
c     Multiple semicolons are reduced to just one semicolon.
c     /* comments are ignored */
c     Any "%" is interpreted as macro-related.
c
c     Note:  the following will trigger an error message and abort execution:
c        (1) infile does not exist
c        (2) outunit is already connected to a file ("bug in program")
c        (3) error reading infile
c        (4) error writing macrofile (or scratch file in current directory)
c        (5) macro syntax error
c        (6) any macro-expanded line exceeds 80 columns
c
c     Constantijn Panis,  26 May 2004
c     Last change: fix nested macros bug in Linux f77.
 
      implicit none
      integer dimstack,maxnummacro
      parameter(dimstack=5,maxnummacro=20)
      integer macrounit(maxnummacro),nummacro,unitstack(dimstack),
     *   comment,string,readunit,inunit,outunit,code,iline,il,
     *   trm,unusedunit
c     integer fgetc
c     integer fgetc
      integer fgetc
      logical exist,opened
      character macroname(maxnummacro)*12,line*128,infile*128,
     *   macrofile*128,c*1,previous*1,ht*1,lf*1,cr*1,sp*1,sq*1,dq*1,
     *   pc*1,sc*1,slash*1,asterisk*1,ascii*1
 
c     ht is the horizontal tab; lf line feed; cr carriage return;
c     sp space; sq '; dq "; pc %; sc semicolon; slash /; asterisk *.
      ht=ascii(9)
      lf=ascii(10)
      cr=ascii(13)
      sp=ascii(32)
      sq=ascii(39)
      dq=ascii(34)
      pc=ascii(37)
      sc=ascii(59)
      slash=ascii(47)
      asterisk=ascii(42)
 
c     Iline keeps track of the current line number in the infile.
      iline=1
      il=0
      previous=ascii(0)
      comment=0
c     String indicates whether the current position is not in a string
c     (0), in a string delimited by single quotes (1), or in a string
c     delimited by double quotes (2).
      string=0
      nummacro=0
 
c     Make sure the file to be stripped exists and that the unit to
c     which the stripped file will be connected is not used yet:
      if(.not.exist(infile))goto 820
      if(opened(outunit))call bug(0,1)
 
c     Open the original control file and a scratch file (in the current
c     directory) for the stripped control file.  Use inunit for
c     the unit of the control file and outunit for its stripped and
c     macro-expanded version.
      inunit=unusedunit()
      open(unit=inunit,file=infile(1:trm(infile)),err=830)
      if(macrofile.eq.' ')then
         open(unit=outunit,status='scratch',err=860)
      else
         open(unit=outunit,file=macrofile,err=860)
      endif
 
c     Start reading the control file, and set the stack of units.
      readunit=inunit
      call zeroint(dimstack,unitstack)
      unitstack(1)=readunit
 
c     Read the input file character-by-character.  Write output file
c     line-by-line, ensuring that it never exceeds 80 characters.  Do
c     not write blank lines.  Replace multiple blanks by a single space.
c     Replace contiguous semicolons by a single semicolon.  If %macro
c     is encountered, write contents of macro (everything until %mend is
c     encountered) to a scratch file.  If another % is encountered,
c     interpret it (until blank, line feed, or semicolon) as a macro
c     and expand.
  100 code=fgetc(readunit,c)
  105 if(code.eq.-1)then
c        End-of-file.  Go to the next file in the stack, if any.
         call popstack(unitstack,dimstack)
         readunit=unitstack(1)
         if(readunit.ne.0)goto 100
c        End-of-file on input file.  If the file did not have a linefeed
c        as its last character, write the last line.  Notepad and Wordpad
c        are known to cause trouble otherwise.  See error messages from
c        Eric Jensen (3/27/2002), Siddhartha Singh (4/7/2002),
c        Benoit Dostie (4/10/2002).
         if(il.ne.0)write(outunit,'(a)')line(1:il)
c        Done; make sure there was no open-ended string or comment.
         if(comment.gt.0.or.string.gt.0)goto 850
c        Clean up by flushing outunit file.
cc       Digital Fortran 90 requires extension
c        call flush(outunit)
cc       SUN f77 requires extension:
c        call flush(outunit)
c        GNU g77 and f77 for Redhat Linux require extension:
         call flush(outunit)
         rewind outunit
         close(inunit)
         do 110 il=1,nummacro
            close(macrounit(il))
  110    continue
         goto 999
      elseif(code.eq.0)then
         if(readunit.eq.inunit.and.c.eq.lf)iline=iline+1
         if(string.eq.0.and.previous.eq.slash.and.c.eq.asterisk)then
c           Start of a comment
            comment=comment+1
c           Don't set previous=asterisk, because /*/ would start and
c           finish the comment, so that /*/* equals *.  Don't leave it
c           as slash, because /** would imply a nested comment.
            previous=sp
            if(comment.eq.1)il=max(0,il-1)
         elseif(comment.gt.0.and.string.eq.0.and.
     *      previous.eq.asterisk.and.c.eq.slash)then
c           End of a comment
            comment=comment-1
c           Don't set previous=slash, because */* would finish and start
c           the comment.  Don't leave it as asterisk, because *// would
c           imply a two closed nested comments.
            previous=sp
         elseif(comment.gt.0)then
c           Inside a comment
            previous=c
         elseif(c.eq.lf)then
c           Digital Fortran terminates all macros with an additional linefeed.
c           Don't write that linefeed.
c           if(readunit.ne.inunit)then
c              code=fgetc(readunit,c)
c              if(code.eq.-1)goto 105
c           endif
c           Write out this line, unless it was blank.
            if(il.eq.0)goto 100
            if(string.ne.0)goto 840
            write(outunit,'(a)')line(1:il)
            il=0
            line=' '
c           if(readunit.ne.inunit)goto 105
         elseif(c.eq.pc)then
c           percent sign; macro definition or macro reference?
            call do_macro(inunit,readunit,macrounit,nummacro,
     *         maxnummacro,dimstack,unitstack,iline,macroname,infile)
         else
c           Record character on line, but replace tabs and multiple
c           spaces by one space, and omit leading blanks.  Replace
c           multiple semicolons by a single semicolon.  Omit
c           carriage returns; they cause problems on UNIX.
            if(c.eq.cr)goto 100
            if(c.eq.ht)c=sp
            if(string.eq.0)then
c              If il=0 Digital Fortran won't accept line(il:il)
               if(c.eq.sp)then
                  if(il.eq.0)goto 100
                  if(line(il:il).eq.' ')goto 100
               endif
               if(c.eq.sc.and.previous.eq.sc)goto 100
               if(c.eq.sq)string=1
               if(c.eq.dq)string=2
            else
               if(string.eq.1.and.c.eq.sq)string=0
               if(string.eq.2.and.c.eq.dq)string=0
            endif
            if(c.ne.sp)previous=c
            il=il+1
            if(il.gt.80)goto 810
            line(il:il)=c
         endif
         goto 100
      else
         goto 800
      endif
 
c     Error encountered while reading control file:
  800 write(6,801)infile(1:trm(infile))
  801 format('ERROR encountered while reading ''',a,
     *   ''' to strip out comments and',/,'expand macros.  Is the ',
     *   'file read-protected?')
      write(6,802)
  802 format('Nested macros are known to trigger this error on Linux ',
     *   'distributions.',/,'Please avoid nested macros under Linux.')
      stop
c     The stripped control file would be wider than 80 columns
  810 call writeaia(6,72,1,'ERROR:  line ',iline,' in '''//
     *   infile(1:trm(infile))//''' exceeds 80 columns (after '//
     *   'expanding macros, if any).')
      write(6,802)
      stop
  820 call not_found(6,infile(1:trm(infile)))
      stop
  830 call wrapstring(6,0,72,0,1,'ERROR trying to read '''//
     *   infile(1:trm(infile))//'''; is the file currently being '//
     *   'accessed by another process?')
      stop
c     String spans multiple lines
  840 call writeaia(6,72,1,'ERROR in '''//infile(1:trm(infile))//
     *   ''': string spans multiple lines around line ',iline-1,'.')
      stop
  850 if(comment.gt.0)write(6,851)'comment',infile(1:trm(infile))
      if(string.gt.0)write(6,851)'string',infile(1:trm(infile))
  851 format(/,'ERROR: ',a,' in ''',a,''' was not closed.')
      stop
  860 if(macrofile.eq.' ')then
         write(6,861)
c 861    format(/,'ERROR opening a scratch data set to pre-process ',
c    *      'macros in the current',/,'working directory.  You ',
c    *      'probably do not have write permission.  Specify',/,
c    *      'command line option "-m \path\filename" to write the ',
c    *      'scratch file',/,'somewhere else.  (You will need to ',
c    *      'remove the scratch file manually.)')
c 861    format(/,'ERROR opening a scratch data set to pre-process ',
c    *      'macros in the current',/,'working directory.  You ',
c    *      'probably do not have write permission.  Specify',/,
c    *      'command line option "-m /path/filename" to write the ',
c    *      'scratch file',/,'somewhere else.  (You will need to ',
c    *      'remove the scratch file manually.)')
  861    format(/,'ERROR opening a scratch data set to pre-process ',
     *      'macros in the current',/,'working directory.  You ',
     *      'probably do not have write permission.  Specify',/,
     *      'command line option "-m /path/filename" to write the ',
     *      'scratch file',/,'somewhere else.  (You will need to ',
     *      'remove the scratch file manually.)')
      else
         write(6,'(1x)')
c        call wrapstring(6,0,72,0,1,'ERROR opening macro pre-process '//
c    *      'file '''//macrofile(1:trm(macrofile))//'''.  You '//
c    *      'probably do not have write permission.  Use command '//
c    *      'line option "-m \path\filename" to specify a filename '//
c    *      'in a directory in which you do have write permission.')
c        call wrapstring(6,0,72,0,1,'ERROR opening macro pre-process '//
c    *      'file '''//macrofile(1:trm(macrofile))//'''.  You '//
c    *      'probably do not have write permission.  Use command '//
c    *      'line option "-m /path/filename" to specify a filename '//
c    *      'in a directory in which you do have write permission.')
         call wrapstring(6,0,72,0,1,'ERROR opening macro pre-process '//
     *      'file '''//macrofile(1:trm(macrofile))//'''.  You '//
     *      'probably do not have write permission.  Use command '//
     *      'line option "-m /path/filename" to specify a filename '//
     *      'in a directory in which you do have write permission.')
      endif
      stop
  999 return
      end
c=========================================================================
      subroutine do_macro(inunit,readunit,macrounit,nummacro,
     *   maxnummacro,dimstack,unitstack,iline,macroname,infile)
 
c     Subroutine to deal with a control file macro: definition or
c     reference.
c
c     readunit    [in/out] unit currently read or to be read (because it
c                       contains a macro definition)
c     macrounit   [out] array with units that are connected to macros
c     nummacro    [in/out] number of macros defined
c     maxnummacro [in]  maximum allowable number of macros
c     dimstack    [in]  dimension of unitstack
c     unitstack   [in/out] stack of integers containing units that
c                       need to be read.  unitstack[1] is currently
c                       being read; if done, go to unitstack[2],
c                       et cetera, until unit zero is encountered;
c                       then all is complete.
c     macroname   [in/out] array with defined macro names
c     infile      [in]  name of to-be-stripped/expanded file.
c
c     Constantijn Panis, 9 July 1996
 
      implicit none
      integer nummacro,maxnummacro,macrounit(maxnummacro),inunit,
     *   readunit,dimstack,unitstack(dimstack),il,iline,n,num,
     *   status,trm,unusedunit,myfputc
c     integer fgetc
c     integer fgetc
      integer fgetc
      character macroname(maxnummacro)*12,infile*(*),c*1,word*12,
     *   ht*1,lf*1,sp*1,sc*1,buffer*5,ascii*1
      logical cmpstri
 
c     ht is horizontal tab; lf line feed; sp space; sc semicolon.
      ht=ascii(9)
      lf=ascii(10)
      sp=ascii(32)
      sc=ascii(59)
 
      il=0
      word=' '
 
c     A % sign has just been encountered.  Read the next word to see
c     whether it is a definition (%macro) or a reference.
  100 if(fgetc(readunit,c).ne.0)goto 800
      if(readunit.eq.inunit.and.c.eq.lf)iline=iline+1
      il=il+1
      if(il.gt.12)goto 810
      word(il:il)=c
      if(cmpstri(word(1:5),'macro'))goto 200
      call char12match(word,macroname,nummacro,n)
      if(n.ne.0)goto 300
      if(c.eq.sp.or.c.eq.lf.or.c.eq.sc)goto 830
      goto 100
 
c     New (renewed?) definition: get the name.
  200 word=sp
      il=0
  210 if(fgetc(readunit,c).ne.0)goto 800
      if(readunit.eq.inunit.and.c.eq.lf)iline=iline+1
      if(c.eq.sp.or.c.eq.lf)goto 210
      if(c.eq.sc)goto 220
      il=il+1
      if(il.gt.12)goto 810
      word(il:il)=c
      goto 210
 
c     Has this name been defined before?
  220 call char12match(word,macroname,nummacro,n)
      if(n.eq.0)goto 230
c     Indeed it has been define before; re-define it.
      n=macrounit(n)
      rewind n
c     For some odd reason (bug in MS Fortran 4.0??), subsequent fputc
c     statements do not erase previously existing macro definitions
c     in this file.  Make sure the file is empty before replacing its
c     contents.
      write(n,'(1x)',err=800)
      rewind n
      goto 240
c     New definition: create scratch file to contain the macro's contents
  230 if(cmpstri(word(1:5),'macro').or.cmpstri(word(1:4),'mend'))
     *   goto 860
      nummacro=nummacro+1
      if(nummacro.gt.maxnummacro)goto 840
      n=unusedunit()
      macrounit(nummacro)=n
      macroname(nummacro)=word
      open(unit=n,status='scratch')
 
c     Read until "%mend" and write exact contents to scratch file
  240 buffer=sp
      num=0
  250 if(fgetc(readunit,c).ne.0)goto 800
      num=num+1
      if(readunit.eq.inunit.and.c.eq.lf)iline=iline+1
      if(num.gt.5)then
         if(myfputc(n,buffer(1:1)).ne.0)goto 820
      endif
      buffer(1:4)=buffer(2:5)
      buffer(5:5)=c
      if(.not.cmpstri(buffer,'%mend'))goto 250
c     End of macro, provided a semicolon follows (somewhat tedious
c     because the semicolon need not immediately follow %mend)
  260 if(fgetc(readunit,c).ne.0)goto 800
      if(readunit.eq.inunit.and.c.eq.lf)iline=iline+1
      if(c.eq.sp.or.c.eq.ht.or.c.eq.lf)goto 260
      if(c.ne.sc)goto 250
c     There appears to be a bug in f77 for Redhat Linux.  Files containing
c     re-defined macros may be empty if they are not Fortran-native read once:
       rewind(n)
       read(n,'(a)',end=999,err=999)c
      goto 999
 
c     Previously defined macro n is being referenced; prepare to read
c     its file.
  300 readunit=macrounit(n)
      rewind readunit
      call pushstack(readunit,unitstack,dimstack,status)
      if(status.gt.0)goto 850
      goto 999
 
  800 call writeline(4,6,0)
      write(6,801)infile(1:trm(infile))
  801 format(/,'ERROR or END-OF-FILE encountered while reading a ',
     *   'macro definition in',/,'''',a,'''.  Did you forget to end ',
     *   'the definition (%mend;) ?')
      stop
  810 call writeline(4,6,0)
      write(6,811)word
  811 format(/,'ERROR: macro name ''%',a,'...'' exceeds 12 characters.')
      stop
  820 call writeline(4,6,0)
      write(6,821)
  821 format(/,'ERROR encountered while attempting to write a scratch ',
     *   'file.',/,'Do you have write permission in the current ',
     *   'directory?')
      stop
  830 call writeline(4,6,0)
      if(il.gt.0)word(il:il)=sp
      if(il.eq.0)write(6,831)
      if(il.gt.0)write(6,832)word(1:trm(word))
      if(cmpstri(word(1:4),'mend'))write(6,833)
  831 format(/,'ERROR: macro name expected after % sign.')
  832 format(/,'ERROR: macro reference ''%',a,''' could not be ',
     *   'resolved.')
  833 format('(Did you attempt to define a macro within another macro ',
     *   'definition?)')
      stop
  840 call writeline(4,6,0)
      call writeaia(6,72,1,'ERROR: the maximum number of '//
     *   'macro definitions (',maxnummacro,') is exceeded.  If '//
     *   'possible, please re-use names of macros that are no '//
     *   'longer used.')
      stop
  850 call writeline(4,6,0)
      call writeaia(6,72,1,'ERROR: the maximum number of '//
     *   'macro reference layers (',dimstack-1,') is exceeded.')
      stop
  860 write(6,861)word(1:trm(word))
  861 format(/,'ERROR: illegal macro name, ''',a,'''.')
  999 return
      end
