      subroutine m3mult4040(task,numrowa,numrowb,numrowc,numcolc,
     *   asparse,b,csparse,d,factor)
 
c     Subroutine to multiply three matrices A*B*C = D.  This routine
c     is similar to m3mult, but exploits special matrix properties.
c     Matrix types can be:
c        0 = generic
c        1 = identity (Only its dimension is stored.)
c        2 = symmetric
c        3 = symmetric and block-diagonal (Only its first block, its
c            dimension, and number of blocks are stored.)
c        4 = sparse in the sense that there is at most one nonzero
c            element in any one row.  (Only its nonzero elements
c            and their column position are stored.)
c
c     In this routine, A is sparse, B generic, C sparse, and D generic.
c     A is stored as two column vectors.  The first contains the column
c     number, the second the actual element.  In other words, A(i,.)=0
c     except A(i,as(i,1)) = as(i,2).  If as(i,1)<=0, all A(i,.)=0.
c     C is stored analogously.
c
c     Conceptual dimensions:
c        A(numrowa,numrowb) -- sparse
c        B(numrowb,numrowc) -- generic
c        C(numrowc,numcolc) -- sparse
c        D(numrowa,numcolc) -- generic
c
c     The four matrices may be in transposed format.  Integer task
c     may be 1000,0100,0010, et cetera, depending on which matrices
c     are transposed.  The first digit refers to A; the second to B;
c     the third to C; the fourth to D.  For example,
c        task = 1100 means that A and B are transposed: A'*B'*C = D
c        task =   10 means that C is transposed: A*B*C' = D
c
c     Dimensions as given above assume no transpositions; if a matrix
c     is transposed, its dimensions really are reversed.
c
c     Integer factor determines:
c        factor = 0     D = A*B*C
c              else     D = D + factor*A*B*C
c
c     Constantijn Panis, 7 March 2002
 
      implicit none
      integer task,numrowa,numrowb,numrowc,numcolc,
     *   factor,i,j,k,jb,jsb,jd,jsd,ib,id,rowd,cold,rowb,colb
c     A is stored as (numrowa,2) if not transposed, (numrowb,2) if transposed
c     C is stored as (numrowc,2) if not transposed, (numcolc,2) if transposed
      double precision asparse(*),b(numrowb*numrowc),
     *   csparse(*),d(numrowa*numcolc),abij,tmp
      logical atranspose,ctranspose
 
c     atranspose = A-transposed?
      atranspose=(mod(task/1000,10).ne.0)
      if(mod(task/100,10).eq.0)then
         jb=1
         jsb=numrowb
      else
         jb=numrowc
         jsb=1
      endif
      ctranspose=(mod(task/10,10).ne.0)
      if(mod(task,10).eq.0)then
         jd=1
         jsd=numrowa
      else
         jd=numcolc
         jsd=1
      endif
 
      if(factor.eq.0)call zerodp(numrowa*numcolc,d)
 
      if((.not.atranspose).and.ctranspose)then
c        Particularly simple case:  D = A*B*C'
         do 100 rowd=1,numrowa
            rowb=int(asparse(rowd))
            if(rowb.le.0)goto 100
            do 110 cold=1,numcolc
               colb=int(csparse(cold))
               if(colb.le.0)goto 110
               ib=1+(rowb-1)*jb+(colb-1)*jsb
               tmp=asparse(numrowa+rowd)*b(ib)*csparse(numcolc+cold)
               if(factor.ne.0.and.factor.ne.1)tmp=dble(factor)*tmp
               id=1+(rowd-1)*jd+(cold-1)*jsd
               d(id) = d(id) + tmp
  110       continue
  100    continue
         goto 999
      endif
 
      do 210 i=1,numrowa
         do 220 j=1,numrowc
            abij = 0.d0
c           Compute AB(i,j)=A(i,.)*B(.,j)
            if(atranspose)then
c              Sparse A is transposed; search for nonzero elements in A(.,i)
c              A(numrowb,numrowa) is stored as asparse(numrowb,2)
               do 230 k=1,numrowb
                  if(int(asparse(k)).eq.i)abij=abij +
     *               asparse(numrowb+k)*b(1+(k-1)*jb+(j-1)*jsb)
  230          continue
            else
c              Sparse A is not transposed; only one element is relevant
c              A(numrowa,numrowb) is stored as asparse(numrowa,2)
               k=int(asparse(i))
               if(k.gt.0)abij=asparse(numrowa+i)*b(1+(k-1)*jb+(j-1)*jsb)
            endif
            if(factor.ne.0.and.factor.ne.1)abij=dble(factor)*abij
c           Accumulate D(i,.) =+ AB(i,j)*C(j,.)
            if(ctranspose)then
c              Sparse C is transposed; search for nonzero elements in C(.,j)
c              C(numrowc,numrowa) is stored as csparse(numcolc,2)
               do 240 k=1,numcolc
c                 D(i,k) =+ AB(i,j)*C(j,k)
                  if(int(csparse(k)).ne.j)goto 240
                  id=1+(i-1)*jd+(k-1)*jsd
                  d(id)=d(id)+abij*csparse(numcolc+k)
  240          continue
            else
c              Sparse C is not transposed; only one element of C(j,.) is nonzero
c              C(numrowa,numrowc) is stored as csparse(numrowc,2)
               k=int(csparse(j))
               if(k.le.0)goto 220
               id=1+(i-1)*jd+(k-1)*jsd
               d(id)=d(id) + abij*csparse(numrowc+j)
            endif
  220    continue
  210 continue
 
  999 return
      end
 
